#define INFO "\
zorkword -- an Infocom(tm) game vocabulary dumper\n\
Copyleft (c) 1991 by Mike Threepoint.  All rights reversed.\n\
Release 7 / Serial number 911103\n\
\n\
Display vocabulary lists from Infocom(tm) adventure game data files.\n\
\n\
Usage:  %s [-1] [-w] [-#] [-f] file...\n\
\n\
\t-1   one-column list\n\
\t-w   wide multiple-column list (default)\n\
\t-#   number vocabulary words\n\
\t-f   print vocabulary flags\n\
\n"

/* Changes in release 7:
 *      Add comments about the file format.  Correct my mistaken assumption
 *      that every entry in the vocabulary table has 3 data bytes -- Sherlock
 *      has only 2, and Shogun has 4.  Add support for non-standard alphabets
 *      in the data file, which Shogun uses.  Make -# and -f ordinary switches
 *      instead of toggles.  Rename program to something more unique than
 *      "vocab".  Show version number of pix data files.
 */

/* Warning:     Using this program on games you haven't solved can result in
 *              giving away passwords, spoiling surprises, and just maybe
 *              relieving the frustration of Dave Lebling's "what verb does
 *              this stupid game want me to use?" puzzles.
 *
 * Disclaimer:  This program works for me, at the moment.  I think it will
 *              work for you.  Nobody within the company told me any technical
 *              details, even when I asked.  I'm just an independent public
 *              domain software author.  If I-Need-Magic doesn't like it, they
 *              are welcome to all zero money I've made on this program.
 *
 * I doubt Infocom's employees will complain, either of them.  Alas, Infocom.
 * I wore a black armband when you went under.  If only you'd stayed solvent.
 * It was my dream to get a job writing adventure games with you.
 *
 * (The closest I've come was maybe writing one function for them.  I hated
 *  using the stupid SETUP.EXE program, which asks you your screen size and
 *  saves it to a 3-byte data file, when I changed my VGA into 50-line mode.
 *  I noticed the Beyond Zork MS-DOS interpreter was in MSC, so I mailed them
 *  a MSC function to get the screen size from the BIOS automatically, and
 *  released the simple code to them.  Some time later, a new text game was
 *  released, with VisiClues.  I started it in 50-line mode, but the screen
 *  was reset to 25-line CGA color mode.  Apparently the screen-length check
 *  is in the wrong place, because then the text ran off the bottom of the
 *  screen and scrolled as if it were still 50 lines long.  I'd mail another
 *  helpful letter, but it's too late now.  Too bad, it would've been great
 *  in a job interview to tell them how I'd been debugging their games.  Sigh.)
 *
 * This code should be portable ANSI C.  You may have trouble using it on a
 * Mac, because the Mac data files contain some system resources before the
 * actual data.  If you use a resource editor to remove them, this program
 * should work on Infocom data files for the Mac, too.
 *
 * For the record, I've compiled this under Turbo C++ in MS-DOS 4.01, and run
 * it on every Infocom data file I have:
 *      A M. F. V.              Hollywood Hijinx        Shogun
 *      Ballyhoo                L. G. of P.             Spellbreaker
 *      Beyond Zork             Lurking Horror          Stationfall
 *      Border Zone             Nord and Bert           Trinity
 *      Bureaucracy             Planetfall              Wishbringer
 *      Deadline                Seastalker              Zork I, II, II
 *      H. G. to the G.         Sherlock                Zork Zero
 *
 * Problems?  Comments?  If you have access to the Internet, e-mail
 * correspondence to linhart@remus.rutgers.edu.
 */

/* Some things to use this program for:
 *    read every last Encyclopedia Frobizzica entry,
 *    be sure you've learned every spell in the Enchanter series,
 *    know all the Wizard of Frobozz's wand's F-words,
 *    discover obscure synonyms (like Glamdring for the Zorkian elvish sword)
 *    learn trivia about internal operations (like the `intnum' noun, used
 *       when you type a number)
 *    play with debugging commands in some games (Stu Galley's especially)
 */

#define MAXCOL  79
#define BUFSIZE 512

#define DEBUG

#if !defined(__STDC__) && !defined(__GNUC__) && !defined(__TURBOC__) && !defined(LINT)
#error This is an ANSI-complaint.  It looks like you are not ANSI-compliant.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#ifndef SEEK_SET
#define SEEK_SET        0
#endif

#define S_BLANK         1
#define S_MACRO1        2
#define S_MACRO2        3
#define S_MACRO3        4
#define S_CAPS          5
#define S_PUNC          6
#define S_FILLER        6
#define S_OFF           7

/* The text compression scheme uses 2 bytes to hold 3 characters of 5 bits each
 * an end-of-text flag bit.
 *
 *    code   meaning
 *    ----   -------
 *       0   space
 *     1-3   next character is an index in macro table 1-3
 *       4   next character is in alphabet 2 (A-Z)
 *       5   next character is in alphabet 3 (special, NL, 0-9, punctuation)
 *    6-31   character in alphabet 1 (a-z), except after codes 1-5
 *
 * The first character in alphabet 3 (the punctuation table) is a special code,
 * which means to compose an 8-bit ASCII character from the last eight bits of
 * the next two 5-bit characters, so any ASCII character left out of the tables
 * (for example, "$") can be represented.  The alphabet 3 code is also used as
 * padding at the end of the string.
 */

/* These are the default alphabets used by the Z-code interpreter, which
 * may be changed by the game.
 */
const char default_alfabets[3][26] = {
        {
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
        'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z'
        },
        {
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'
        },
        {
        '\0' /* ASCII literal */, '\n',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        '.', ',', '!', '?', '_', '#', '\'', '\"', '/', '\\', '-', ':', '(', ')'
        }
};

#define ERR_CHAR '?'

char alfabets[3][26];

char *macros[3][32] = { NULL };

typedef unsigned char   byte;
typedef unsigned short  word;
typedef unsigned short  z_word;

struct info_header {
        byte    z_version;      /* 3=Standard 4=Plus 5=VisiClues 6=Graphics */
        byte    flags;          /* x01=Header error, x02=Time: status */
        z_word  release;
        z_word  resident_size;
        z_word  game_offset;
        z_word  vocab_offset;
        z_word  object_offset;
        z_word  variable_offset;
        z_word  save_area_size;
        z_word  script_status;  /* x01=script on, x02=no proportional */
        char    rev_date[6];
        z_word  macro_offset;
        z_word  verify_length;  /* in words */
        z_word  verify_checksum;
        z_word  reserved1[8];
        z_word  fkey_offset;
        z_word  reserved2[2];
        z_word  alfabet_offset;
        z_word  reserved3[5];
};

FILE    *infile;
short   column = 0,
        columns = 0;
char    show_flags = 0,
        numbers = 0,
#ifdef DEBUG
        debug_info = 0,
#endif
        did_file = 0;
char *  prog_name = "zorkword";

unsigned
unzword ( z_word z )
/* change a most-significant-byte-first 2-byte integer into a 2-byte integer */
{
        byte *zp = (byte *)&z;
        return (zp[0] << 8) + zp[1];
}

void
newline ( void )
{
        putchar('\n');
        column = 0;
}

short
end_of_text ( z_word chars )
/* is the end of text bit set on this compressed text word? */
{
        return chars >> 15;
}

char *
expand ( word chars )
/* break a compressed text word into three characters */
{
        static char buf[4] = {0, 0, 0, 0};

        buf[2] = (chars & 0x1F) + 1;
        chars >>= 5;
        buf[1] = (chars & 0x1F) + 1;
        chars >>= 5;
        buf[0] = (chars & 0x1F) + 1;

        return buf;
}

char *
decode ( char *s )
/* translate the 5-bit characters into text */
{
        int             len = strlen(s);
        static char     new[BUFSIZE];
        unsigned        newlen = 0;

        /* trim off code 5's used as padding at the end of the string */
        while (s[len-1] == S_FILLER)
                s[--len] = '\0';

        while (*s) {
                switch (*s) {
                        case S_MACRO1:
                        case S_MACRO2:
                        case S_MACRO3:
                                /* next character is index in macro table */
                                /* shouldn't appear in vocabulary list */
                                {
                                        char *tmp = macros[(*s)-2][(*(s+1))-1];
                                        s++;
                                        if (tmp) {
                                                strcpy(&new [ newlen ], tmp);
                                                newlen += strlen(tmp);
                                        }
                                }
                                break;
                        case S_CAPS:
                                /* next character is in capitals alphabet */
                                if (*(s+1) >= S_OFF)
                                        new [ newlen++ ] = alfabets[1][*++s - S_OFF];
                                else
                                        new [ newlen++ ] = ERR_CHAR;
                                break;
                        case S_PUNC:
                                /* next character is in punctuation alphabet */
                                if (*(s+1) >= S_OFF)
                                        if (*++s == S_OFF) {
                                                /* combine next 2 codes into ASCII character */
                                                new [ newlen ] = ((*++s - 1) & 0x03) << 5;
                                                new [ newlen++ ] += *++s - 1;
                                        } else
                                                new [ newlen++ ] = alfabets[2][*s - S_OFF];
                                else
                                        new [ newlen++ ] = ERR_CHAR;
                                break;
                        case S_BLANK:
                                new [ newlen++ ] = ' ';
                                break;
                        default:
                                new [ newlen++ ] = alfabets[0][*s - S_OFF];
                }
                s++;
        }

        new [ newlen ] = '\0';

        return new;
}

void
disp_ch ( char x )
{
        putchar(x);
        column++;
}

void
disp_str ( char *fmt, ... )
{
        va_list         argptr;
        static char     buf[16];
        short           len;

        va_start(argptr, fmt);
        vsprintf(buf, fmt, argptr);
        va_end(argptr);

        column += printf(buf);
}

void
disp_bits ( char c )
/* display character in binary */
{
        unsigned b;

        disp_ch(' ');
        for (b = 0x80; b; b >>= 1)
                disp_ch(c & b ? '1' : '0');
}

void
error ( char *fmt, ... )
{
        va_list         argptr;

        fprintf(stderr, "\nError: ");

        va_start(argptr, fmt);
        vfprintf(stderr, fmt, argptr);
        va_end(argptr);

        exit(1);
}

void
read_error ( void )
{
        error("Can't read file at offset %04lX.\n", ftell(infile));
}

void
seek_pos ( unsigned long pos )
{
        if (fseek(infile, pos, SEEK_SET) != 0)
                error("Can't seek offset %04lX.\n", pos);
}

/* At the object table offset recorded in the file header is a table of
 * data for each object in the game, preceded by a number of extra
 * records.
 *
 * In version 3 games:
 *      byte    attributes[3];
 *      byte    location,       [parent in object tree]
 *              next,           [sibling in object tree]
 *              contents;       [child in object tree]
 *      z_word  data_offset;    [file offset for rest of data]
 *      byte    reserved;
 *
 * In later games:
 *      byte    attributes[6];
 *      z_word  location,
 *              next,
 *              contents;
 *      z_word  data_offset;
 *
 * At the data offset, the first bytes are:
 *      1 byte          number of compressed text words to follow
 *      n words         compressed text name of the object
 */

/* At the vocabulary table offset recorded in the file header is this data:
 *      1 byte          number of punctuation characters to follow
 *      n characters    characters recognized by the parser as punctuation
 *      1 byte          length of each vocabulary entry (usually 7 or 9)
 *      1 word          number of vocabulary entries in the table
 *
 * Then the vocabulary entries themselves.  The first 2 or 3 words of each
 * entry are the compressed text, followed by a few data bytes associated
 * with each entry.
 *
 * Version 3 games use an entry length of 7, for 2 compressed text words
 * encoding 6 letters, and 3 data bytes.  Later games have larger entry
 * lengths, for 3 compressed text words encoding 9 letters, and 2 or more
 * data bytes (but usually 3).
 */
void
dump_vocab ( unsigned long pos )
{
        register unsigned       count = 0, index;
        word                    words;
        byte                    letters_per_word,
                                zwords_per_word,
                                flag_bytes;
        short                   vocab_entry_size,
                                entry_width,
                                entries_per_line;
        char                    format[sizeof("%%-%ds")];
        int                     punct;
        char *                  buf;
        byte *                  flags;

#ifdef DEBUG
        printf("Vocabulary table at offset %04lX\n", pos);
#endif

        seek_pos(pos);

        /* display the list of punctuation characters before the table */
        if ((punct = getc(infile)) == EOF)
                read_error();
        if (punct) {
                printf("Punctuation: ");
                for (; punct; punct--)
                        putchar(getc(infile));
                putchar('\n');
        }

        /* size of each vocabulary entry */
        if ((vocab_entry_size = getc(infile)) == EOF)
                read_error();

        /* number of entries */
        if (fread(&words, sizeof(words), 1, infile) < 1)
                read_error();
        words = unzword(words);
        if (!numbers)
                printf("%u vocabulary entries\n", words);

        /* determine where each word ends and the data begins */
        zwords_per_word = vocab_entry_size <= 7 ? 2 : 3;
        letters_per_word = zwords_per_word * 3;
        flag_bytes = vocab_entry_size - zwords_per_word * 2;

        /* figure out how wide the columns are */
        entry_width = letters_per_word + 2;
        if (numbers)
                entry_width += 5;
        if (show_flags)
                entry_width += flag_bytes * (columns == 1 ? 9 : 3) + 1;
        entries_per_line = columns ? columns : (MAXCOL + 2) / entry_width;

        /* set up buffers */
        buf = malloc(letters_per_word + 1);
        flags = malloc(flag_bytes);
        sprintf(format, "%%-%ds", letters_per_word);

        while ( count < words ) {
                ++count;
                if (numbers)
                        disp_str("%04d ", count);

                /* decode the text for this vocabulary entry */
                for (index = 0; index < zwords_per_word; index++) {
                        word    z;

                        if (fread(&z, sizeof(z), 1, infile) < 1)
                                read_error();
                        z = unzword(z);
                        if (index)
                                strcat(buf, expand(z));
                        else
                                strcpy(buf, expand(z));
                }

                disp_str(format, decode(buf));

                /* get the additional data bytes for this entry */
                if (fread(flags, sizeof(char), flag_bytes, infile) < flag_bytes)
                        read_error();

                if (show_flags) {
                        register short n;

                        disp_ch(' ');
                        for (n = 1; n <= flag_bytes; n++)
                                if (columns == 1)
                                        disp_bits(flags[n]);
                                else
                                        disp_str(" %02x", flags[n]);
                }

                /* if another column won't fit, go to the next line */
                if (entries_per_line > 1 && (count % entries_per_line))
                        disp_str("  ");
                else
                        newline();
        }

        free(buf);
        free(flags);

        if (column)
                newline();
}

/* At the macro offset recorded in the file header are 3 tables of 32 words
 * each containing the file offsets of the text for macro tables 1-3.
 */
#ifdef DEBUG
void
read_macros ( unsigned long pos )
{
        register short  table, elem;
        z_word          macro_offsets[3][32];

        seek_pos(pos);

        fread(macro_offsets, sizeof(z_word), 3 * 32, infile);

        for (table = 0; table < 3; table++)
                for (elem = 0; elem < 32; elem++) {
                        macros[table][elem] = NULL;

                        /* display macro trivia */
                        if (debug_info)
                                printf("Macro %d-%02d: ", table, elem);

                        if ((pos = unzword(macro_offsets[table][elem])) != 0) {
                                char    macro[BUFSIZE];
                                z_word  z;
                                char    *text;

                                seek_pos(pos *= sizeof(z_word));

                                macro[0] = '\0';
                                do {
                                        if (fread(&z, sizeof(z), 1, infile) < 1)
                                                read_error();
                                        z = unzword(z);
                                        if (macro[0])
                                                strcat(macro, expand(z));
                                        else
                                                strcpy(macro, expand(z));
                                } while (!end_of_text(z));

                                text = decode(macro);
                                macros[table][elem] = malloc(strlen(text) + 1);
                                strcpy(macros[table][elem], text);

                                if (debug_info)
                                        printf("\"%s\"\n", text);
                        }
                }
}
#endif

/* At the alphabet offset recorded in the file header are 3 tables of 26
 * characters containing the text characters for alphabets 1-3.
 */
void
read_alfabets ( unsigned long pos )
{
        register short  table, elem;

        /* if included in the file, read it */
        if (pos) {
                seek_pos(pos);
                fread(alfabets, 1, 3 * 26, infile);

#ifdef DEBUG
                if (debug_info)
                        for (table = 0; table < 3; table++) {
                                printf("Alphabet %d: ", table+1);
                                for (elem = 0; elem < 26; elem++)
                                        putchar(isprint(alfabets[table][elem]) ?
 alfabets[table][elem] : ERR_CHAR);
                                putchar('\n');
                        }
#endif

                return;
        }

        /* load default alfabets */
        for (table = 0; table < 3; table++)
                for (elem = 0; elem < 26; elem++)
                        alfabets[table][elem] = default_alfabets[table][elem];

}

void
frob_file ( const char *filename )
{
        struct info_header      header;
        int                     index;

        if((infile = fopen(filename, "rb")) == NULL)
                error("Can't open file \"%s\".\n", filename);

        printf("%s:\n", filename);

        if (fread(&header, sizeof(header), 1, infile) < 1)
                read_error();

#ifdef __TURBOC__
        /* I accidentally ran this on one too many .EXE files. */
        if (header.z_version == 'M' && header.flags == 'Z') {
                puts("MS-DOS Interpreter");
                fclose(infile);
                return;
        }
#endif

        /* Catch pix data files */
        if (header.z_version == 1) {
                z_word  release;

                seek_pos(0x0D);
                fread(&release, sizeof(z_word), 1, infile);

                printf("Pix release %u\n", unzword(release));

                fclose(infile);
                return;
        }

        /* display release number and date from the file header */
        printf("Release %u / Serial number %.6s\n", unzword(header.release), header.rev_date);

#ifdef DEBUG
        if (debug_info)
                printf("Z code version %d\n", header.z_version);
#endif

        /* load alphabets and macros */
        read_alfabets(unzword(header.alfabet_offset));

#ifdef DEBUG
        if (debug_info)
                read_macros(unzword(header.macro_offset));
#endif

        dump_vocab(unzword(header.vocab_offset));

        fclose(infile);
}

#ifndef LINT
const char sccsid[] = "@(#) " __FILE__ " by Mike Threepoint compiled " __DATE__;
#endif

void
info ( void )
{
        printf(INFO, prog_name);
        exit(0);
}

void
parse_opt ( char p )
{
        switch (p) {
                case 'w':
                        columns = 0;
                        break;
                case '#':
                case 'n':
                        numbers = 1;
                        break;
                case 'f':
                case 'b':
                        show_flags = 1;
                        break;
#ifdef DEBUG
                case 'd':
                case 'm':
                        debug_info = 1;
                        break;
#endif
                case 'h':
                case '?':
                        info();
                        break;
                default:
                        /* if it's a digit, use that many columns */
                        if (isdigit(p))
                                columns = (p - '0');
        }
}

void
parse ( char *parm )
{
        switch (*parm) {
                case '/':
                        parse_opt(*++parm);
                        break;
                case '-':
                        while (*++parm) parse_opt(*parm);
                        break;
                default:
                        if (did_file) newline();
                        frob_file(parm);
                        did_file = 1;
        }
}

int
main ( const unsigned argc, char *argv[] )
{
        if (argv[0]) prog_name = argv[0];

        if (argc > 1) {
                register count;

                if (strcmp(argv[1], "?") == 0)
                        info();

                for (count = 1; count < argc; count++)
                        parse(argv[count]);

                if (did_file) return 0;
        }

        info();
        return 1;
}
